@@ -156,7 +156,7 @@
 //! [`Builder`]: struct.Builder.html
 //! [`Error`]: ../struct.Error.html
 
-use {SendStream, RecvStream, ReleaseCapacity};
+use {SendStream, RecvStream, ReleaseCapacity, PingPong};
 use codec::{Codec, RecvError, SendError, UserError};
 use frame::{Headers, Pseudo, Reason, Settings, StreamId};
 use proto;
@@ -187,10 +187,10 @@ use std::usize;
 ///
 /// [module]: index.html
 #[must_use = "futures do nothing unless polled"]
-pub struct Handshake<T, B: IntoBuf = Bytes> {
+pub struct Handshake<T, B = Bytes> {
     builder: Builder,
     inner: WriteAll<T, &'static [u8]>,
-    _marker: PhantomData<B>,
+    _marker: PhantomData<fn(B)>,
 }
 
 /// Initializes new HTTP/2.0 streams on a connection by sending a request.
@@ -319,31 +319,13 @@ pub struct PushPromise {
     response: PushedResponseFuture,
 }
 
-#[derive(Debug)]
 /// A stream of pushed responses and corresponding promised requests
+#[derive(Debug)]
+#[must_use = "streams do nothing unless polled"]
 pub struct PushPromises {
     inner: proto::OpaqueStreamRef,
 }
 
-impl Stream for PushPromises {
-    type Item = PushPromise;
-    type Error = ::Error;
-
-    fn poll(&mut self) -> Poll<Option<Self::Item>, Self::Error> {
-        match try_ready!(self.inner.poll_pushed()) {
-            Some((request, response)) => {
-                let response = PushedResponseFuture {
-                    inner: ResponseFuture {
-                        inner: response, push_promise_consumed: false
-                    }
-                };
-                Ok(Async::Ready(Some(PushPromise{request, response})))
-            }
-            None => Ok(Async::Ready(None)),
-        }
-    }
-}
-
 /// Builds client connections with custom configuration values.
 ///
 /// Methods can be chained in order to set the configuration values.
@@ -879,6 +861,14 @@ impl Builder {
         self
     }
 
+    /// Sets the maximum table size of the dynamic header decoder.
+    ///
+    /// By default, this value is 4,096 bytes.
+    pub fn header_table_size(&mut self, max: u32) -> &mut Self {
+        self.settings.set_header_table_size(Some(max));
+        self
+    }
+
     /// Sets the maximum number of concurrent streams.
     ///
     /// The maximum concurrent streams setting only controls the maximum number
@@ -1282,6 +1272,17 @@ where
         assert!(size <= proto::MAX_WINDOW_SIZE);
         self.inner.set_target_window_size(size);
     }
+
+    /// Takes a `PingPong` instance from the connection.
+    ///
+    /// # Note
+    ///
+    /// This may only be called once. Calling multiple times will return `None`.
+    pub fn ping_pong(&mut self) -> Option<PingPong> {
+        self.inner
+            .take_user_pings()
+            .map(PingPong::new)
+    }
 }
 
 impl<T, B> Future for Connection<T, B>
@@ -1340,6 +1341,12 @@ where
             codec.set_max_recv_header_list_size(max as usize);
         }
 
+        // The server receives the settings first, so we can just assume
+        // they will be ACKed, or else bad server.
+        if let Some(max) = self.builder.settings.header_table_size() {
+            codec.set_max_recv_header_table_size(max as usize);
+        }
+
         // Send initial settings frame
         codec
             .buffer(self.builder.settings.clone().into())
@@ -1416,6 +1423,27 @@ impl ResponseFuture {
     }
 }
 
+// ===== impl PushPromises =====
+
+impl Stream for PushPromises {
+    type Item = PushPromise;
+    type Error = ::Error;
+
+    fn poll(&mut self) -> Poll<Option<Self::Item>, Self::Error> {
+        match try_ready!(self.inner.poll_pushed()) {
+            Some((request, response)) => {
+                let response = PushedResponseFuture {
+                    inner: ResponseFuture {
+                        inner: response, push_promise_consumed: false
+                    }
+                };
+                Ok(Async::Ready(Some(PushPromise{request, response})))
+            }
+            None => Ok(Async::Ready(None)),
+        }
+    }
+}
+
 // ===== impl PushPromise =====
 
 impl PushPromise {
