@@ -1,4 +1,4 @@
-use super::{StreamDependency, StreamId};
+use super::{util, StreamDependency, StreamId};
 use frame::{Error, Frame, Head, Kind};
 use hpack;
 
@@ -12,6 +12,10 @@ use string::String;
 use std::fmt;
 use std::io::Cursor;
 
+// Minimum MAX_FRAME_SIZE is 16kb, so save some arbitrary space for frame
+// head and other header bits.
+const MAX_HEADER_LENGTH: usize = 1024 * 16 - 100;
+
 /// Header frame
 ///
 /// This could be either a request or a response.
@@ -232,6 +236,10 @@ impl Headers {
         self.header_block.is_over_size
     }
 
+    pub(crate) fn has_too_big_field(&self) -> bool {
+        self.header_block.has_too_big_field()
+    }
+
     pub fn into_parts(self) -> (Pseudo, HeaderMap) {
         (self.header_block.pseudo, self.header_block.fields)
     }
@@ -274,12 +282,17 @@ impl<T> From<Headers> for Frame<T> {
 
 impl fmt::Debug for Headers {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        f.debug_struct("Headers")
+        let mut builder = f.debug_struct("Headers");
+        builder
             .field("stream_id", &self.stream_id)
-            .field("stream_dep", &self.stream_dep)
-            .field("flags", &self.flags)
-            // `fields` and `pseudo` purposefully not included
-            .finish()
+            .field("flags", &self.flags);
+
+        if let Some(ref dep) = self.stream_dep {
+            builder.field("stream_dep", dep);
+        }
+
+        // `fields` and `pseudo` purposefully not included
+        builder.finish()
     }
 }
 
@@ -656,11 +669,11 @@ impl From<HeadersFlag> for u8 {
 
 impl fmt::Debug for HeadersFlag {
     fn fmt(&self, fmt: &mut fmt::Formatter) -> fmt::Result {
-        fmt.debug_struct("HeadersFlag")
-            .field("end_stream", &self.is_end_stream())
-            .field("end_headers", &self.is_end_headers())
-            .field("padded", &self.is_padded())
-            .field("priority", &self.is_priority())
+        util::debug_flags(fmt, self.0)
+            .flag_if(self.is_end_headers(), "END_HEADERS")
+            .flag_if(self.is_end_stream(), "END_STREAM")
+            .flag_if(self.is_padded(), "PADDED")
+            .flag_if(self.is_priority(), "PRIORITY")
             .finish()
     }
 }
@@ -704,9 +717,9 @@ impl From<PushPromiseFlag> for u8 {
 
 impl fmt::Debug for PushPromiseFlag {
     fn fmt(&self, fmt: &mut fmt::Formatter) -> fmt::Result {
-        fmt.debug_struct("PushPromiseFlag")
-            .field("end_headers", &self.is_end_headers())
-            .field("padded", &self.is_padded())
+        util::debug_flags(fmt, self.0)
+            .flag_if(self.is_end_headers(), "END_HEADERS")
+            .flag_if(self.is_padded(), "PADDED")
             .finish()
     }
 }
@@ -839,6 +852,46 @@ impl HeaderBlock {
             .map(|(name, value)| decoded_header_size(name.as_str().len(), value.len()))
             .sum::<usize>()
     }
+
+    /// Iterate over all pseudos and headers to see if any individual pair
+    /// would be too large to encode.
+    pub(crate) fn has_too_big_field(&self) -> bool {
+        macro_rules! pseudo_size {
+            ($name:ident) => ({
+                self.pseudo
+                    .$name
+                    .as_ref()
+                    .map(|m| decoded_header_size(stringify!($name).len() + 1, m.as_str().len()))
+                    .unwrap_or(0)
+            });
+        }
+
+        if pseudo_size!(method) > MAX_HEADER_LENGTH {
+            return true;
+        }
+
+        if pseudo_size!(scheme) > MAX_HEADER_LENGTH {
+            return true;
+        }
+
+        if pseudo_size!(authority) > MAX_HEADER_LENGTH {
+            return true;
+        }
+
+        if pseudo_size!(path) > MAX_HEADER_LENGTH {
+            return true;
+        }
+
+        // skip :status, its never going to be too big
+
+        for (name, value) in &self.fields {
+            if decoded_header_size(name.as_str().len(), value.len()) > MAX_HEADER_LENGTH {
+                return true;
+            }
+        }
+
+        false
+    }
 }
 
 fn decoded_header_size(name: usize, value: usize) -> usize {
