@@ -140,17 +140,39 @@ where
 
         let key = match me.store.find_entry(id) {
             Entry::Occupied(e) => e.key(),
-            Entry::Vacant(e) => match me.actions.recv.open(id, Open::Headers, &mut me.counts)? {
-                Some(stream_id) => {
-                    let stream = Stream::new(
-                        stream_id,
-                        me.actions.send.init_window_sz(),
-                        me.actions.recv.init_window_sz(),
-                    );
+            Entry::Vacant(e) => {
+                // Client: it's possible to send a request, and then send
+                // a RST_STREAM while the response HEADERS were in transit.
+                //
+                // Server: we can't reset a stream before having received
+                // the request headers, so don't allow.
+                if !P::is_server() {
+                    // This may be response headers for a stream we've already
+                    // forgotten about...
+                    if me.actions.may_have_forgotten_stream::<P>(id) {
+                        debug!(
+                            "recv_headers for old stream={:?}, sending STREAM_CLOSED",
+                            id,
+                        );
+                        return Err(RecvError::Stream {
+                            id,
+                            reason: Reason::STREAM_CLOSED,
+                        });
+                    }
+                }
 
-                    e.insert(stream)
-                },
-                None => return Ok(()),
+                match me.actions.recv.open(id, Open::Headers, &mut me.counts)? {
+                    Some(stream_id) => {
+                        let stream = Stream::new(
+                            stream_id,
+                            me.actions.send.init_window_sz(),
+                            me.actions.recv.init_window_sz(),
+                        );
+
+                        e.insert(stream)
+                    },
+                    None => return Ok(()),
+                }
             },
         };
 
@@ -180,8 +202,9 @@ where
                     Ok(()) => Ok(()),
                     Err(RecvHeaderBlockError::Oversize(resp)) => {
                         if let Some(resp) = resp {
-                            let _ = actions.send.send_headers(
+                            let sent = actions.send.send_headers(
                                 resp, send_buffer, stream, counts, &mut actions.task);
+                            debug_assert!(sent.is_ok(), "oversize response should not fail");
 
                             actions.send.schedule_implicit_reset(
                                 stream,
@@ -203,8 +226,13 @@ where
                 }
             } else {
                 if !frame.is_end_stream() {
-                    // TODO: Is this the right error
-                    return Err(RecvError::Connection(Reason::PROTOCOL_ERROR));
+                    // Receiving trailers that don't set EOS is a "malformed"
+                    // message. Malformed messages are a stream error.
+                    proto_err!(stream: "recv_headers: trailers frame was not EOS; stream={:?}", stream.id);
+                    return Err(RecvError::Stream {
+                        id: stream.id,
+                        reason: Reason::PROTOCOL_ERROR,
+                    });
                 }
 
                 actions.recv.recv_trailers(frame, stream)
@@ -230,7 +258,26 @@ where
                     return Ok(());
                 }
 
-                trace!("recv_data; stream not found: {:?}", id);
+                if me.actions.may_have_forgotten_stream::<P>(id) {
+                    debug!(
+                        "recv_data for old stream={:?}, sending STREAM_CLOSED",
+                        id,
+                    );
+
+                    let sz = frame.payload().len();
+                    // This should have been enforced at the codec::FramedRead layer, so
+                    // this is just a sanity check.
+                    assert!(sz <= super::MAX_WINDOW_SIZE as usize);
+                    let sz = sz as WindowSize;
+
+                    me.actions.recv.ignore_data(sz)?;
+                    return Err(RecvError::Stream {
+                        id,
+                        reason: Reason::STREAM_CLOSED,
+                    });
+                }
+
+                proto_err!(conn: "recv_data: stream not found; id={:?}", id);
                 return Err(RecvError::Connection(Reason::PROTOCOL_ERROR));
             },
         };
@@ -260,6 +307,7 @@ where
         let id = frame.stream_id();
 
         if id.is_zero() {
+            proto_err!(conn: "recv_reset: invalid stream ID 0");
             return Err(RecvError::Connection(Reason::PROTOCOL_ERROR));
         }
 
@@ -342,6 +390,10 @@ where
             // the value they send in the last stream identifier, since the
             // peers might already have retried unprocessed requests on another
             // connection.")
+            proto_err!(conn:
+                "recv_go_away: last_stream_id ({:?}) > max_stream_id ({:?})",
+                last_stream_id, actions.recv.max_stream_id(),
+            );
             return Err(RecvError::Connection(Reason::PROTOCOL_ERROR));
         }
 
@@ -426,7 +478,10 @@ where
                 stream.state.ensure_recv_open()?;
                 stream.key()
             }
-            None => return Err(RecvError::Connection(Reason::PROTOCOL_ERROR)),
+            None => {
+                proto_err!(conn: "recv_push_promise: initiating stream is in an invalid state");
+                return Err(RecvError::Connection(Reason::PROTOCOL_ERROR))
+            },
         };
 
         // TODO: Streams in the reserved states do not count towards the concurrency
@@ -576,7 +631,7 @@ where
         // prioritize layer. If prioritization reorders new streams, this
         // implicitly closes the earlier stream IDs.
         //
-        // See: carllerche/h2#11
+        // See: hyperium/h2#11
         let mut me = self.inner.lock().unwrap();
         let me = &mut *me;
 
@@ -660,26 +715,17 @@ where
 
         let key = match me.store.find_entry(id) {
             Entry::Occupied(e) => e.key(),
-            Entry::Vacant(e) => match me.actions.recv.open(id, Open::Headers, &mut me.counts) {
-                Ok(Some(stream_id)) => {
-                    let stream = Stream::new(stream_id, 0, 0);
+            Entry::Vacant(e) => {
+                let stream = Stream::new(id, 0, 0);
 
-                    e.insert(stream)
-                },
-                _ => return,
+                e.insert(stream)
             },
         };
 
         let stream = me.store.resolve(key);
-        let actions = &mut me.actions;
         let mut send_buffer = self.send_buffer.inner.lock().unwrap();
         let send_buffer = &mut *send_buffer;
-
-        me.counts.transition(stream, |counts, stream| {
-            actions.send.send_reset(
-                reason, send_buffer, stream, counts, &mut actions.task);
-            actions.recv.enqueue_reset_expiration(stream, counts)
-        })
+        me.actions.send_reset(stream, reason, &mut me.counts, send_buffer);
     }
 
     pub fn send_go_away(&mut self, last_processed_id: StreamId) {
@@ -752,11 +798,17 @@ where
         Ok(())
     }
 
+    #[cfg(feature = "unstable")]
     pub fn num_active_streams(&self) -> usize {
         let me = self.inner.lock().unwrap();
         me.store.num_active_streams()
     }
 
+    pub fn has_streams(&self) -> bool {
+        let me = self.inner.lock().unwrap();
+        me.counts.has_streams()
+    }
+
     pub fn has_streams_or_other_references(&self) -> bool {
         let me = self.inner.lock().unwrap();
         me.counts.has_streams() || me.refs > 1
@@ -848,14 +900,10 @@ impl<B> StreamRef<B> {
         let me = &mut *me;
 
         let stream = me.store.resolve(self.opaque.key);
-        let actions = &mut me.actions;
         let mut send_buffer = self.send_buffer.inner.lock().unwrap();
         let send_buffer = &mut *send_buffer;
 
-        me.counts.transition(stream, |counts, stream| {
-            actions.send.send_reset(
-                reason, send_buffer, stream, counts, &mut actions.task)
-        })
+        me.actions.send_reset(stream, reason, &mut me.counts, send_buffer);
     }
 
     pub fn send_response(
@@ -1047,6 +1095,17 @@ impl OpaqueStreamRef {
             .release_capacity(capacity, &mut stream, &mut me.actions.task)
     }
 
+    pub(crate) fn clear_recv_buffer(&mut self) {
+        let mut me = self.inner.lock().unwrap();
+        let me = &mut *me;
+
+        let mut stream = me.store.resolve(self.key);
+
+        me.actions
+            .recv
+            .clear_recv_buffer(&mut stream);
+    }
+
     pub fn stream_id(&self) -> StreamId {
         self.inner.lock()
             .unwrap()
@@ -1178,6 +1237,22 @@ impl<B> SendBuffer<B> {
 // ===== impl Actions =====
 
 impl Actions {
+    fn send_reset<B>(
+        &mut self,
+        stream: store::Ptr,
+        reason: Reason,
+        counts: &mut Counts,
+        send_buffer: &mut Buffer<Frame<B>>,
+    ) {
+        counts.transition(stream, |counts, stream| {
+            self.send.send_reset(
+                reason, send_buffer, stream, counts, &mut self.task);
+            self.recv.enqueue_reset_expiration(stream, counts);
+            // if a RecvStream is parked, ensure it's notified
+            stream.notify_recv();
+        });
+    }
+
     fn reset_on_recv_stream_err<B>(
         &mut self,
         buffer: &mut Buffer<Frame<B>>,
@@ -1213,6 +1288,26 @@ impl Actions {
         }
     }
 
+    /// Check if we possibly could have processed and since forgotten this stream.
+    ///
+    /// If we send a RST_STREAM for a stream, we will eventually "forget" about
+    /// the stream to free up memory. It's possible that the remote peer had
+    /// frames in-flight, and by the time we receive them, our own state is
+    /// gone. We *could* tear everything down by sending a GOAWAY, but it
+    /// is more likely to be latency/memory constraints that caused this,
+    /// and not a bad actor. So be less catastrophic, the spec allows
+    /// us to send another RST_STREAM of STREAM_CLOSED.
+    fn may_have_forgotten_stream<P: Peer>(&self, id: StreamId) -> bool {
+        if id.is_zero() {
+            return false;
+        }
+        if P::is_local_init(id) {
+            self.send.may_have_created_stream(id)
+        } else {
+            self.recv.may_have_created_stream(id)
+        }
+    }
+
     fn clear_queues(&mut self,
                     clear_pending_accept: bool,
                     store: &mut Store,
